@@ -13,11 +13,12 @@ from django import template
 
 from django.core.serializers.json import DjangoJSONEncoder
 from django.db import models
-from django.urls import reverse
+from django.urls import reverse, is_valid_path
 from django.utils.encoding import force_text
 from django.utils.functional import Promise
 from django.utils.html import format_html
 from django.utils.safestring import mark_safe
+from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
 
 register = template.Library()
 
@@ -26,6 +27,9 @@ from django.utils.translation import gettext_lazy as _
 
 if PY_VER != '2':
     from importlib import reload
+    from urllib.parse import parse_qsl
+else:
+    from urlparse import parse_qsl
 
 
 def unicode_to_str(u, encoding='utf-8'):
@@ -150,6 +154,11 @@ def get_config(key):
     return __get_config(key)
 
 
+@register.filter
+def get_value(value):
+    return value
+
+
 @register.simple_tag
 def get_version():
     return simpleui.get_version()
@@ -248,14 +257,7 @@ def menus(context, _get_config=None):
 
     # 给每个菜单增加一个唯一标识，用于tab页判断
     eid = 1000
-    for i in data:
-        eid += 1
-        i['eid'] = eid
-        if 'models' in i:
-            for k in i.get('models'):
-                eid += 1
-                k['eid'] = eid
-
+    handler_eid(data, eid)
     menus_string = json.dumps(data, cls=LazyEncoder)
 
     # 把data放入session中，其他地方可以调用
@@ -265,6 +267,15 @@ def menus(context, _get_config=None):
     return '<script type="text/javascript">var menus={}</script>'.format(menus_string)
 
 
+def handler_eid(data, eid):
+    for i in data:
+        eid += 1
+        i['eid'] = eid
+        if 'models' in i:
+            eid = handler_eid(i.get('models'), eid)
+    return eid
+
+
 def get_icon(obj, name=None):
     temp = get_config_icon(name)
     if temp != '':
@@ -458,9 +469,24 @@ def simple_version():
 @register.simple_tag(takes_context=True)
 def get_model_url(context):
     # reverse()
-    opts = context.get('opts')
-    key = 'admin:{}_{}_changelist'.format(opts.app_label, opts.model_name)
-    return reverse(key)
+    opts = context.get("opts")
+    key = "admin:{}_{}_changelist".format(opts.app_label, opts.model_name)
+    url = reverse(key)
+    preserved_filters = dict(parse_qsl(context.get("preserved_filters")))
+    if "_changelist_filters" in preserved_filters:
+        preserved_filters = preserved_filters["_changelist_filters"]
+        url = add_preserved_filters({"preserved_filters": preserved_filters, "opts": opts}, url)
+    return url
+
+
+@register.simple_tag(takes_context=True)
+def get_model_ajax_url(context):
+    opts = context.get("opts")
+    key = "admin:{}_{}_ajax".format(opts.app_label, opts.model_name)
+    try:
+        return reverse(key)
+    except:
+        pass
 
 
 @register.simple_tag
@@ -487,9 +513,14 @@ def get_boolean_choices():
 
 @register.simple_tag(takes_context=True)
 def get_previous_url(context):
-    return context.request.META.get('HTTP_REFERER')
+    referer = context.request.META.get("HTTP_REFERER")
+    if not referer or context.request.META.get("PATH_INFO") in referer:
+        # return to model list
+        return get_model_url(context)
+    return context.request.META.get("HTTP_REFERER")
 
 
 @register.simple_tag(takes_context=True)
 def get_verbose_name_plural(context):
     return context['cl'].model._meta.verbose_name_plural
+
